#include "stdio.h"
#include <windows.h>
#include "types.h"
#include "dvd_drive.h"
#include "disk.h"
#include "general.h"
#include "unscramble.h"
#include "constants.h"
#include "byteorder.h"

/* Cache always deals with 16-sector blocks. All numbers refer to the 16-sector blocks */
#define DISC_MINIMUM_CACHE_SIZE 5
#define DISC_DEFAULT_CACHE_SIZE 40
#define CACHE_ENTRY_INVALID ((U32) -1)

#define DEFAULT_READ_METHOD     4
#define MAX_READ_RETRIES        5

#define DISC_GAMECUBE_SECTORS_NO 0x0AE0B0	/* 712880 */
#define DISC_WII_SECTORS_NO_SL 0x230480 	/* 2294912 */
#define DISC_WII_SECTORS_NO_DL 0x3F69C0 	/* 4155840 */

static char disc_type_strings[4][15] = 
{
	"GameCube",
	"Wii",
	"Wii_DL",
	"Unknown"
};

static char disc_region_strings[4][15] =
{
	"Europe/PAL",
	"USA/NTSC",
	"Japan/NTSC",
	"Unknown"
};

/* The following list has been derived from GCMtool, http://www.sadistech.com/gcmtool/, GPL2, whose data come from YAGCD */
static struct {
	char *code;
	char *name;
} makers[] = {
	{"01", "Nintendo"},
	{"08", "Capcom"},
	{"4F", "Eidos"},
	{"51", "Acclaim"},
	{"5D", "Midway"},
	{"5G", "Hudson"},
	{"64", "LucasArts"},
	{"69", "Electronic Arts"},
	{"6S", "TDK Mediactive"},
	{"8P", "Sega"},
	{"AF", "Namco"},
	{"DA", "Tomy"},
	{"EM", "Konami"},
	{"52", "Activision"},
	{"B2", "Bandai"},
	{"41", "Ubisoft"},
	{"A4", "Mirage Studios"},
	{NULL, NULL}
};

static void disc_set_unscrambling (DISK_TYPE *d, U8 unscramble);
static U8 disc_set_read_method (DISK_TYPE *d, S32 method);
static S32 disc_read_sector_1 (DISK_TYPE *d, U32 sector_no, U8 **data, U8 **rawdata);
static S32 disc_read_sector_2 (DISK_TYPE *d, U32 sector_no, U8 **data, U8 **rawdata);
static S32 disc_read_sector_3 (DISK_TYPE *d, U32 sector_no, U8 **data, U8 **rawdata);
static S32 disc_read_sector_4 (DISK_TYPE *d, U32 sector_no, U8 **data, U8 **rawdata);
static void disc_cache_init (DISK_TYPE *d, U32 size);
static DISK_TYPE_ENUM disc_detect_type (DISK_TYPE *d);
static S32 disc_check_seeds (DISK_TYPE *d);
static U8 disc_cache_lookup_block (DISK_TYPE *d, U32 block, U8 **data, U8 **rawdata);
static void disc_cache_add_block (DISK_TYPE *d, U32 block, U8 *data, U8 *rawdata);
static U8 disc_analyze (DISK_TYPE *d);
static U8 disc_check_update (DISK_TYPE *d);
static void disc_cache_destroy (DISK_TYPE *d);
static void disc_find_gameid(DISK_TYPE *d);

/**
 * Creates a new structure representing a Nintendo GameCube/Wii optical disc.
 * @param dvd_device The CD/DVD-ROM device, in OS-dependent format (i.e.: /dev/something on Unix, x: on Windows).
 * @return The newly-created structure, to be used with the other commands.
 */
DISK_TYPE *disc_new (char *dvd_device)
{
	DVD_DRIVE_TYPE *dvd;
	DISK_TYPE *d;

	if ((dvd = dvd_drive_new (dvd_device)))
    {
		d = (DISK_TYPE *) malloc (sizeof (DISK_TYPE));
		memset (d, 0, sizeof (DISK_TYPE));
		d -> dvd = dvd;
		unscramble_init();
		disc_set_unscrambling (d, TRUE);	/* Unscramble by default */
		disc_set_read_method (d, DEFAULT_READ_METHOD);
		disc_cache_init (d, DISC_DEFAULT_CACHE_SIZE);
	}
    else
    {
		d = NULL;
	}

	return (d);
}

/**
 * Sets the disc read method.
 * @param d The disc structure.
 * @param method The requested method.
 * @return True if the method was set correctly, false otherwise (i. e.: method too small/big).
 */
static U8 disc_set_read_method (DISK_TYPE *d, S32 method)
{
	U8 out;

	out = TRUE;

    d->read_method = method;

	switch((d->read_method))
    {
		case 1:
			d -> read_sector_fun_ptr = disc_read_sector_1;
			break;
		case 2:
			d -> read_sector_fun_ptr = disc_read_sector_2;
			break;
		case 3:
			d -> read_sector_fun_ptr = disc_read_sector_3;
			break;
		case 4:
			d -> read_sector_fun_ptr = disc_read_sector_4;
			break;
		default:
			out = FALSE;
			d -> read_method = -1;
			d -> read_sector_fun_ptr = NULL;
			break;
	}

	if(TRUE == out)
    {
        #ifdef DEBUG_MODE
		fprintf (stderr,"Read method set to %d\n", method);
        #endif
	}
    else
    {
		printf("Cannot set read method\n");
	}

	return (out);
}

static S32 disc_read_sector_1 (DISK_TYPE *d, U32 sector_no, U8 **data, U8 **rawdata)
{
	U8 out;
	U32 start_block;
	S32 ret, retry;
	U8 buf[RAW_BLOCK_SIZE];
	U8 buf_unscrambled[BLOCK_SIZE];
	
	start_block = sector_no / SECTORS_PER_BLOCK;

	out = FALSE;

	for (retry = 0; !out && retry < MAX_READ_RETRIES; retry++)
    {
		/* Assume everything will turn out well */
		out = TRUE;

		if (retry > 0)
        {
            #ifdef DEBUG_MODE
			fprintf (stderr,"Read retry %d for sector %u ", retry, sector_no);
            #endif

			/* Try to reset in-memory data by seeking to a distant sector */
			if (sector_no > 1000)
            {
				dvd_read_sector_dummy (d -> dvd, 0, NULL);
            }
			else
            {
				dvd_read_sector_dummy (d -> dvd, 1500, NULL);
            }
		}

		if ((ret = dvd_read_sector_dummy (d -> dvd, sector_no, NULL)) >= 0)
        {
			if(dvd_memdump(d -> dvd, 0, 1, RAW_BLOCK_SIZE, buf) < 0)
            {
				printf("Memdump failed");
				out = FALSE;
				retry = MAX_READ_RETRIES;		/* Well, if this fails going on is useless */
			} 
            else
            {
				if (d -> unscrambling)
                {
					/* Try to unscramble all data to see if EDC fails */
					if (0 == unscramble_16sectors(sector_no, buf, buf_unscrambled))
                    {
						out = FALSE;
                    }
				}
			}

			if (out)
            {
				/* If data were unscrambled correctly, add them to the cache */
				disc_cache_add_block (d, start_block, buf_unscrambled, buf);
			}
		}
        else
        {
			printf ("dvd_read_sector_dummy() failed with %d", ret);
			out = FALSE;
		}
	}

	if (!out)
    {
        #ifdef DEBUG_MODE
		fprintf (stderr,"Too many disk read errors, giving up\n");
        #endif
    }

	return (out);
}

static S32 disc_read_sector_2 (DISK_TYPE *d, U32 sector_no, U8 **data, U8 **rawdata)
{
	U8 out;
	U32 start_block;
	S32 j, ret, retry;
	U8 buf[4][16 * 2064];
	U8 buf_unscrambled[4][16 * 2048];

	start_block = sector_no / SECTORS_PER_BLOCK;

	out = FALSE;

	for (retry = 0; !out && retry < MAX_READ_RETRIES; retry++)
    {
		/* Assume everything will turn out well */
		out = TRUE;

		if (retry > 0)
        {
            #ifdef DEBUG_MODE
			fprintf (stderr,"Read retry %d for sector %u", retry, sector_no);
            #endif

			/* Try to reset in-memory data by seeking to a distant sector */
			if (sector_no > 1000)
            {
				dvd_read_sector_streaming (d -> dvd, 0, NULL, NULL, 0);
            }
			else
            {
				dvd_read_sector_streaming (d -> dvd, 1500, NULL, NULL, 0);
            }
		}

        if((ret = dvd_read_sector_streaming (d -> dvd, sector_no, NULL, NULL, 0)) >= 0)
        {
    		if ((ret = dvd_read_sector_streaming (d -> dvd, sector_no, NULL, NULL, 0)) >= 0)
            {
    			for (j = 0; j < 4 && sector_no + j * 16 < d -> sectors_no && out; j++)
                {
    				/* Dumping in a single block is faster */
                    if (dvd_memdump (d -> dvd, 0 + (j * 16 * 2064), 1, 16 * 2064, buf[j]) < 0) 
                    {
    					printf("Memdump failed");
    					out = FALSE;
    					retry = MAX_READ_RETRIES;		/* Well, if this fails going on is useless */
    				}
                    else 
                    {
    					if (d -> unscrambling)
                        {
    
    						/* Try to unscramble all data to see if EDC fails */
    						if (0 == unscramble_16sectors (sector_no + (j * 16), buf[j], buf_unscrambled[j]))
                            {
    							out = FALSE;
                            }
    					}
    				}
    
    			}
    
    			if(out) 
                {
    				/* It seems all data was unscrambled correctly, so cache them out */
    				for (j = 0; j < 4 && sector_no + j * 16 < d -> sectors_no; j++)
                    {
    					disc_cache_add_block (d, start_block + j, buf_unscrambled[j], buf[j]);
                    }
    			}
    		}
            else
            {
    			printf("dvd_read_sector_streaming() failed with %d", ret);
    			out = FALSE;
    		}
        }
        else
        {
			printf ("dvd_read_sector_dummy() failed with %d", ret);
			out = FALSE;
		}
	}

	if(!out)
    {
        #ifdef DEBUG_MODE
		fprintf(stderr,"Too many disk read errors, giving up\n");
        #endif
    }

	return (out);
}

static S32 disc_read_sector_3 (DISK_TYPE *d, U32 sector_no, U8 **data, U8 **rawdata)
{
	U8 out;
	U32 ram_offset;
	S32 j, k, ret, retry;
	U8 *sect, buf[4][RAW_BLOCK_SIZE];
	U8 readbuf[BLOCK_SIZE];
	U8 buf_unscrambled[4][BLOCK_SIZE];
	U32 start_block;

	start_block = sector_no / SECTORS_PER_BLOCK;
	
	out = FALSE;

	for (retry = 0; !out && retry < MAX_READ_RETRIES; retry++)
    {
		/* Assume everything will turn out well */
		out = TRUE;

		if (retry > 0)
        {
            #ifdef DEBUG_MODE
			fprintf (stderr,"Read retry %d for sector %u", retry, sector_no);
            #endif

			/* Try to reset in-memory data by seeking to a distant sector */
			if (sector_no > 1000)
            {
				dvd_read_sector_streaming (d -> dvd, 0, NULL, NULL, 0);
            }
			else
            {
				dvd_read_sector_streaming (d -> dvd, 1500, NULL, NULL, 0);
            }
		}

		/* First READ command, this will cache 5 16-sector blocks. Immediately dump relevant data */
		if (sector_no > d -> sectors_no - 1000)
        {
			dvd_read_sector_streaming (d -> dvd, sector_no - 16 * 5 * 2, NULL, NULL, 0);
        }
		else
        {
			dvd_read_sector_streaming (d -> dvd, sector_no + 16 * 5, NULL, NULL, 0);
        }

		if ((ret = dvd_read_sector_streaming (d -> dvd, sector_no, NULL, readbuf, sizeof (readbuf))) >= 0)
        {
			for (j = 0; j < 4 && sector_no + j * 16 < d -> sectors_no && out; j++)
            {
				/* Reconstruct raw sectors */
				for (k = 0; k < 16; k++)
                {
					sect = &buf[j][k * RAW_SECTOR_SIZE];
					ram_offset = (j * RAW_BLOCK_SIZE) + k * RAW_SECTOR_SIZE;
					/* Get first 12 bytes (ID. IED and CPR_MAI fields) and last 4 bytes (EDC field) with memdump */
					if (dvd_memdump (d -> dvd, ram_offset, 1, 12, sect) < 0)
                    {
						printf("Memdump (1) failed");
						out = FALSE;
						retry = MAX_READ_RETRIES;		/* Well, if this fails going on is useless */
					}
                    else if (dvd_memdump (d -> dvd, ram_offset + 2060, 1, 4, sect + 2060) < 0)
                    {	/* Dumping in a single block is faster */
						printf("Memdump (2) failed");
						out = FALSE;
					}
				}
			}

			/* Now the same for remaining 4 16-sector blocks */
			for (j = 0; j < 4 && sector_no + j * 16 < d -> sectors_no && out; j++)
            {
				if (j == 0 || (ret = dvd_read_sector_streaming (d -> dvd, sector_no + j * 16, NULL, readbuf, sizeof (readbuf))) >= 0)
                {
					/* Copy "user data" field which has been incorrectly unscrambled by the DVD drive firmware */
					for (k = 0; k < 16; k++)
                    {
						sect = &buf[j][k * RAW_SECTOR_SIZE];
						memcpy (sect + 12, readbuf + k * SECTOR_SIZE, SECTOR_SIZE);
					}

					if (d -> unscrambling)
                    {
                        /* re-scramble with dvd seed */
                        rescramble_16sectors(sector_no + (j * 16), buf[j]);

						/* now unscramble correctly all data to see if EDC fails */
						if (0 == unscramble_16sectors (sector_no + (j * 16), buf[j], buf_unscrambled[j]))
                        {
							out = FALSE;
                        }
					}
				}
                else
                {
					printf("dvd_read_sector_streaming() failed with %d", ret);
					out = FALSE;
				}
			}

			if (out)
            {
				/* It seems all data were unscrambled correctly, so cache them out */
				for (j = 0; j < 4 && sector_no + j * SECTORS_PER_BLOCK < d -> sectors_no; j++)
                {
					disc_cache_add_block (d, start_block + j, buf_unscrambled[j], buf[j]);
                }
			}
		} 
        else
        {
			printf("dvd_read_sector_streaming() failed with %d", ret);
			out = FALSE;
		}
	}

	if(!out)
    {
        #ifdef DEBUG_MODE
		fprintf(stderr,"Too many retries, giving up\n");
        #endif
    }

	return (out);
}

static S32 disc_read_sector_4 (DISK_TYPE *d, U32 sector_no, U8 **data, U8 **rawdata)
{
	U8 out;
	U32 ram_offset;
	S32 j, k, ret, retry;
	U8 *sect, buf[4][RAW_BLOCK_SIZE];
	U8 readbuf[BLOCK_SIZE], tmp[16];
	U8 buf_unscrambled[4][BLOCK_SIZE];
	U32 start_block;

	start_block = sector_no / SECTORS_PER_BLOCK;

	out = FALSE;

	for (retry = 0; !out && retry < MAX_READ_RETRIES; retry++)
    {
		/* Assume everything will turn out well */
		out = TRUE;

		if (retry > 0)
        {
            #ifdef DEBUG_MODE
			fprintf(stderr,"Read retry %d for sector %u", retry, sector_no);
            #endif

			/* Try to reset in-memory data by seeking to a distant sector */
			if (sector_no > 1000)
            {
				dvd_read_sector_streaming (d -> dvd, 0, NULL, NULL, 0);
            }
			else
            {
				dvd_read_sector_streaming (d -> dvd, 1500, NULL, NULL, 0);
            }
		}

		/* First READ command, this will cache 5 16-sector blocks. Immediately dump relevant data */
		if (sector_no > d -> sectors_no - 1000)
        {
			dvd_read_sector_streaming (d -> dvd, sector_no - 16 * 5 * 2, NULL, NULL, 0);
        }
		else
        {
			dvd_read_sector_streaming (d -> dvd, sector_no + 16 * 5, NULL, NULL, 0);
        }

		if ((ret = dvd_read_sector_streaming (d -> dvd, sector_no, NULL, readbuf, BLOCK_SIZE)) >= 0)
        {
			for (j = 0; j < 4 && sector_no + j * 16 < d -> sectors_no && out; j++)
            {
				/* Reconstruct raw sectors */
				for (k = 0; k < 16; k++)
                {
					sect = &buf[j][k * RAW_SECTOR_SIZE];
					ram_offset = (j * RAW_BLOCK_SIZE) + k * RAW_SECTOR_SIZE;
					/* Get first 12 bytes (ID. IED and CPR_MAI fields) and last 4 bytes (EDC field) with memdump */
					if (j == 0 && k == 0)
                    {
						if (dvd_memdump (d -> dvd, ram_offset, 1, 12, sect) < 0)
                        {
							printf("Memdump (1) failed");
							out = FALSE;
							retry = MAX_READ_RETRIES;		/* Well, if this fails going on is useless */
						}
					}
                    else
                    {
						memcpy (sect, tmp + 4, 12);
					}
					
                    /* Dumping in a single block is faster */
					if (out && dvd_memdump (d -> dvd, ram_offset + 2060, 1, 16, tmp) < 0) 
                    {	
						printf("Memdump (2) failed");
						out = FALSE;
					}
                    else
                    {
						memcpy (sect + 2060, tmp, 4);
					}
				}
			}

			/* Now the same for remaining 4 16-sector blocks */
			for (j = 0; j < 4 && sector_no + j * 16 < d -> sectors_no && out; j++)
            {
				if (j == 0 || (ret = dvd_read_sector_streaming (d -> dvd, sector_no + j * 16, NULL, readbuf, BLOCK_SIZE)) >= 0)
                {
					/* Copy "user data" field which has been incorrectly unscrambled by the DVD drive firmware */
					for (k = 0; k < 16; k++) 
                    {
						sect = &buf[j][k * RAW_SECTOR_SIZE];
						memcpy (sect + 12, readbuf + k * SECTOR_SIZE, SECTOR_SIZE);
					}

					if (d -> unscrambling)
                    {

                        /* re-scramble with dvd seed */
                        rescramble_16sectors(sector_no + (j * 16), buf[j]);

						/* Try to unscramble all data to see if EDC fails */
						if (0 == unscramble_16sectors (sector_no + (j * 16), buf[j], buf_unscrambled[j]))
                        {
							out = FALSE;
                        }
					}

				}
                else
                {
					printf("dvd_read_sector_streaming() failed with %d", ret);
					out = FALSE;
				}
			}

			if (out)
            {
				/* It seems all data were unscrambled correctly, so cache them out */
				for (j = 0; j < 4 && sector_no + j * SECTORS_PER_BLOCK < d -> sectors_no; j++)
                {
					disc_cache_add_block (d, start_block + j, buf_unscrambled[j], buf[j]);
                }
			}
		}
        else
        {
			printf("dvd_read_sector_streaming() failed with %d", ret);
			out = FALSE;
		}
	}

	if (!out)
    {
        #ifdef DEBUG_MODE
		fprintf(stderr,"Too many retries, giving up\n");
        #endif
    }

	return (out);
}


/**
 * Controls the unscrambling process.
 * @param d The disc structure.
 * @param unscramble If true, every raw sectors read will be unscrambled to check if they are error-free, otherwise read data will be returned as-is.
 */
static void disc_set_unscrambling (DISK_TYPE *d, U8 unscramble)
{
	d->unscrambling = unscramble;
    #ifdef DEBUG_MODE
	fprintf (stderr,"Sectors unscrambling %s\n", unscramble ? "enabled" : "disabled");
    #endif

	return;
}

static void disc_cache_init (DISK_TYPE *d, U32 size)
{
	U32 i;

	if (size < DISC_MINIMUM_CACHE_SIZE)
    {
		printf("Invalid cache size %u (must be >= %u)", size, DISC_MINIMUM_CACHE_SIZE);
		exit (3);
	}
    else
    {
		d -> cache_size = size;
		d -> cache = (U8 **) malloc (sizeof (U8 *) * size);
		d -> raw_cache = (U8 **) malloc (sizeof (U8 *) * size);

		for (i = 0; i < size; i++)
        {
			d -> cache[i] = (U8 *) malloc (sizeof (U8) * BLOCK_SIZE);
			d -> raw_cache[i] = (U8 *) malloc (sizeof (U8) * RAW_BLOCK_SIZE);
		}

		d -> cache_map = (U32 *) malloc (sizeof (U32) * size);

		for (i = 0; i < size; i++)
        {
			d -> cache_map[i] = CACHE_ENTRY_INVALID;
        }
	}

	return;
}

static void disc_cache_add_block (DISK_TYPE *d, U32 block, U8 *data, U8 *rawdata)
{
	U32 pos;
	
	pos = block % d -> cache_size;

	memcpy (d -> cache[pos], data, BLOCK_SIZE);
	memcpy (d -> raw_cache[pos], rawdata, RAW_BLOCK_SIZE);
	d -> cache_map[pos] = block;

    #ifdef DEBUG_MODE
	fprintf(stderr,"Cached block %u (sectors %u-%u) at position %u", block, block * SECTORS_PER_BLOCK, (block + 1) * SECTORS_PER_BLOCK - 1, pos);
    #endif

	return;
}

U8 disc_init (DISK_TYPE *d)
{
	U8 out = TRUE;
	
	d -> sectors_no = 1000;
	disc_detect_type (d);
	out = disc_check_seeds(d);

    if(TRUE == out)
    {
	    if(disc_analyze(d))
        {
		    disc_check_update (d);
		    out = TRUE;
	    }
        else
        {
		    out = FALSE;
	    }
    }

	return (out);
}

static S32 disc_check_seeds (DISK_TYPE *d)
{
	S32 i = 0;
    S32 out = TRUE;

    /* first find gameID checksum */
    disc_find_gameid(d);

	/* As a Nintendo GameCube/Wii disc should not have too many keys, 20 should be enough */
    printf("testing cal'ed disc seeds...");   

	for (i = 0; i < 20 * 16; i += 16)
    {
		out = disc_read_sector (d, i, NULL, NULL);
        if(FALSE == out)
        {
            break;
        }
    }

	return(out);
}

/* I cant work out the per game id checksum so this brute forces it */
static void disc_find_gameid(DISK_TYPE *d)
{
    U8 i;
    S32 out = FALSE;

    /* brute force gameid checksum (value can be 0 to 15)*/
    for(i = 0; i < 16; i++)
    {
        /* set the gameid checksum */
        unscramble_set_gameid(i);
        /* try and unscramble sector 16 with it */
        out = disc_read_sector (d, 16, NULL, NULL);
        /* if it unscrambled then it s the correct gameid checksum */
        if(TRUE == out)
        {
            printf("Using per gameID checksum of 0x%02x\n",i);
            break;
        }
    }

    if(FALSE == out)
    {
        printf("Cant find gameID checksum!\n");
    }
}

/* wiidevel@stacktic.org */
static U8 disc_check_update (DISK_TYPE *d)
{
	U8 *buf;
	U32 x;
	U8 unscramble_old;

	if (d -> type == DISC_TYPE_WII || d -> type == DISC_TYPE_WII_DL)
    {
		/* Force unscrambling for this read */
		unscramble_old = d -> unscrambling;
		disc_set_unscrambling (d, TRUE);

		/* We need to read offset 0x50004 of the disc. Sector 160 has offset 0x50000 */
		if (disc_read_sector (d, 160, &buf, NULL))
        {
			x = my_ntohl (*(U32 *) (buf + 4));

			if (x == 0xA5BED6AE)
            {
				d -> has_update = FALSE;
            }
			else
            {
				d -> has_update = TRUE;
            }
		}
        else
        {
			printf("disc_check_update() failed");
		}

		disc_set_unscrambling (d, unscramble_old);
	}
    else
    {
		/* GameCube discs never have an update, as actually the GC firmware cannot be upgrade */
		d -> has_update = FALSE;
	}

	return (d -> has_update);
}

/**
 * Reads a sector from the disc (or from the cache), using the preset read method.
 * @param d The disc structure.
 * @param sector_no The requested sector number.
 * @param data A buffer to hold the unscrambled sector data (or NULL).
 * @param rawdata A buffer to hold the raw sector data (or NULL).
 * @return 
 */
S32 disc_read_sector (DISK_TYPE *d, U32 sector_no, U8 **data, U8 **rawdata)
{
	U32 block;
	U8 *cdata;
    U8 *crawdata;
	S32 out;

	/* Unscrambled data cannot be requested if unscrambling was disabled */
	MY_ASSERT (!(data && !d -> unscrambling));
	
	block = sector_no / SECTORS_PER_BLOCK;
	
	/* See if sector is in cache */
	if (!(out = disc_cache_lookup_block (d, block, &cdata, &crawdata)))
    {
		/* Requested block is not in cache, try to read it from media */
		out = d->read_sector_fun_ptr(d, sector_no, data, rawdata);
		
		/* Now requested sector is in cache, for sure ;) */
		if (out)
        {
			MY_ASSERT (disc_cache_lookup_block (d, block, &cdata, &crawdata));
        }
	}

	if (out)
    {
		if (data)
        {
			*data = cdata + (sector_no % SECTORS_PER_BLOCK) * SECTOR_SIZE;
        }

		if (rawdata)
        {
			*rawdata = crawdata + (sector_no % SECTORS_PER_BLOCK) * RAW_SECTOR_SIZE;
        }
	}
    else
    {
		if (data)
        {
			*data = NULL;
        }

		if (rawdata)
        {
			*rawdata = NULL;
        }
	}
		
	return (out);
}

static U8 disc_analyze (DISK_TYPE *d)
{
	U8 *buf;
	char tmp[0x03E0 + 1];
	U8 unscramble_old;
    U8 out;

	/* Force unscrambling for this read */
	unscramble_old = d -> unscrambling;
	disc_set_unscrambling (d, TRUE);
	
	if (disc_read_sector (d, 0, &buf, NULL))
    {
		/* System ID */
		d -> system_id = buf[0];

		/* Game ID */
		strncpy (d -> game_id, (char *) buf + 1, 2);
		d -> game_id[2] = '\0';

		/* Region */
		switch (buf[3]) {
			case 'P':
				d -> region = DISC_REGION_PAL;
				break;
			case 'E':
				d -> region = DISC_REGION_NTSC;
				break;
			case 'J':
				d -> region = DISC_REGION_JAPAN;
				break;
			default:
				d -> region = DISC_REGION_UNKNOWN;
				break;
		}

		/* Maker code */
		strncpy (d -> maker, (char *) buf + 4, 2);
		d -> maker[2] = '\0';

		/* Version */
		d -> version = buf[7];
		snprintf (tmp, sizeof (tmp), "1.%02u", d -> version);
		d -> version_string = strdup(tmp);

		/* Game title */
		memcpy (tmp, buf + 0x0020, sizeof (tmp) - 1);
		tmp[sizeof (tmp) - 1] = '\0';
		strtrimr (tmp);
		d -> title = strdup(tmp);

		out = TRUE;
	}
    else
    {
		printf("Cannot analyze disc");
		out = FALSE;
	}

	disc_set_unscrambling (d, unscramble_old);

	return (out);
}

/**
 * Frees resources used by a disc structure and destroys it.
 * @param d The disc structure.
 * @return NULL.
 */
void *disc_destroy (DISK_TYPE *d)
{
	disc_cache_destroy (d);
	free(d -> version_string);
	free(d -> title);
	dvd_drive_destroy (d -> dvd);
	free(d);

	return (NULL);
}

static void disc_cache_destroy (DISK_TYPE *d)
{
	U32 i;
	
	free(d -> cache_map);
	
	for (i = 0; i < d -> cache_size; i++)
    {
	    free (d -> cache[i]);
		free (d -> raw_cache[i]);
	}

    free(d -> cache);
	free(d -> raw_cache);
	d -> cache_size = 0;

	return;
}

/**
 * Retrieves the disc type.
 * @param d The disc structure.
 * @param dt This will be set to the disc type.
 * @param dt_s This will point to a string describing the disc type.
 * @return A string describing the disc type.
 */
char *disc_get_type (DISK_TYPE *d, DISK_TYPE_ENUM *dt, char **dt_s)
{
	if (dt)
    {
		*dt = d -> type;
    }

	if (dt_s)
    {
		if (d -> type < DISC_TYPE_UNKNOWN)
        {
			*dt_s = disc_type_strings[d -> type];
        }
		else
        {
			*dt_s = disc_type_strings[DISC_TYPE_UNKNOWN];
        }
	}

	return (*dt_s);
}

/**
 * Retrieves the disc game ID.
 * @param d The disc structure.
 * @param gid_s This will point to a string containing the game ID.
 * @return A string containing the game ID.
 */
char *disc_get_gameid (DISK_TYPE *d, char **gid_s)
{
	if (gid_s)
    {
		*gid_s = d -> game_id;
    }

	return (*gid_s);
}

/**
 * Retrieves the disc region.
 * @param d The disc structure.
 * @param dr This will be set to the disc region.
 * @param dr_s This will point to a string describing the disc region.
 * @return A string describing the disc region.
 */
char *disc_get_region (DISK_TYPE *d, DISK_REGION_ENUM *dr, char **dr_s)
{
	if (dr)
    {
		*dr = d -> region;
    }
	
	if (dr_s)
    {
		if (d -> region < DISC_REGION_UNKNOWN)
        {
			*dr_s = disc_region_strings[d -> region];
        }
		else
        {
			*dr_s = disc_region_strings[DISC_REGION_UNKNOWN];
        }
	}

	return (*dr_s);
}

/**
 * Retrieves the disk maker.
 * @param d The disc structure.
 * @param m This will point to a string containing the disc maker ID.
 * @param m_s This will point to a string describing the disc maker.
 * @return A string describing the disc maker.
 */
char *disc_get_maker (DISK_TYPE *d, char **m, char **m_s)
{
	U32 i;
	
	if (m)
    {
		*m = d -> maker;
    }

	if (m_s)
    {
		for (i = 0; makers[i].code; i++)
        {
			if (strcasecmp (d -> maker, makers[i].code) == 0)
            {
				*m_s = makers[i].name;
				break;
			}
		}

		if (!makers[i].code)
        {
			*m_s = "Unknown";
		}
	}

	return (*m_s);
}

/**
 * Retrieves the disc version.
 * @param d The disc structure.
 * @param v This will contain the version ID.
 * @param v_s This will point to a string describing the disc version.
 * @return A string describing the disc version.
 */
char *disc_get_version (DISK_TYPE *d, U8 *v, char **v_s)
{
	if (v)
    {
		*v = d -> version;
    }
	
	if (v_s)
    {
		*v_s = d -> version_string;
    }

	return (*v_s);
}

/**
 * Retrieves the disc game title.
 * @param d The disc structure.
 * @param t_s This will point to a string describing the disc title.
 * @return A string describing the disc title.
 */
char *disc_get_title (DISK_TYPE *d, char **t_s)
{
	if (t_s)
    {
		*t_s = d -> title;
    }

	return (*t_s);
}

static U8 disc_cache_lookup_block (DISK_TYPE *d, U32 block, U8 **data, U8 **rawdata)
{
	U32 pos;
	U8 out;

	pos = block % d -> cache_size;

	if (d -> cache_map[pos] == block)
    {
		#ifdef DEBUG_MODE
	    fprintf(stderr,"Cache HIT for block %u", block);
        #endif

		if (data)
        {
			*data = d -> cache[pos];
        }

		if (rawdata)
        {
			*rawdata = d -> raw_cache[pos];
        }

		out = TRUE;
	}
    else
    {
	    #ifdef DEBUG_MODE
	    fprintf(stderr,"Cache MISS for block %u", block);
        #endif

		if (data)
        {
			*data = NULL;
        }

		if (rawdata)
        {
			*rawdata = NULL;
        }

		out = FALSE;
	}

	return (out);
}

/* We could also use the 'System ID' (first byte of the image) to tell the discs apart */
static DISK_TYPE_ENUM disc_detect_type (DISK_TYPE *d)
{
	REQ_SENSE_TYPE sense;

	/* Try to read a sector beyond the end of GameCube discs */
	if((0 == dvd_read_sector_dummy (d -> dvd, DISC_GAMECUBE_SECTORS_NO + 100, &sense)) && (sense.sense_key == 0x05) && (sense.asc == 0x21))
    {
		d -> type = DISC_TYPE_GAMECUBE;
		d -> sectors_no = DISC_GAMECUBE_SECTORS_NO;
	} 
    else
    {
		if ((0 == dvd_read_sector_dummy (d -> dvd, DISC_WII_SECTORS_NO_SL + 100, &sense)) && (sense.sense_key == 0x05) && (sense.asc == 0x21))
        {
			d -> type = DISC_TYPE_WII;
			d -> sectors_no = DISC_WII_SECTORS_NO_SL;
		}
        else
        {
			d -> type = DISC_TYPE_WII_DL;
			d -> sectors_no = DISC_WII_SECTORS_NO_DL;
		}
	}

	return (d -> type);
}

char *disc_get_drive_model_string (DISK_TYPE *d)
{
	return (dvd_get_model_string (d -> dvd));
}


U8 disc_get_drive_support_status (DISK_TYPE *d)
{
	return (dvd_get_support_status (d -> dvd));
}

/**
 * Retrieves the number of sectors of the disc.
 * @param d The disc structure.
 * @return The number of sectors.
 */
U32 disc_get_sectors_no (DISK_TYPE *d) 
{
	return (d -> sectors_no);
}
