#include <stdio.h>
#include <string.h>
#include <windows.h>
#include <WinIoCtl.h>
#include "ntddscsi.h"
#include "types.h"
#include "dvd_drive.h"
#include "general.h"
#include "hitachi.h"

#define MMC_CMD_TIMEOUT 10

static S32 dvd_get_drive_info (DVD_DRIVE_TYPE *dvd);
static void dvd_assign_functions (DVD_DRIVE_TYPE *dvd);

/**
 * Creates a new structure representing a CD/DVD-ROM drive.
 * @param device The CD/DVD-ROM device, in OS-dependent format (i.e.: /dev/something on Unix, x: on Windows).
 * @return The newly-created structure, to be used with the other commands, or NULL if the drive could not be initialized.
 */
DVD_DRIVE_TYPE *dvd_drive_new (char *device)
{
	DVD_DRIVE_TYPE *dvd;
	HANDLE fd;
	char dev[40];
	
    #ifdef DEBUG_MODE
	fprintf(stderr, "Trying to open DVD device %s\n", device);
    #endif

	sprintf(dev, "\\\\.\\%c:", device[0]);

	if ((fd = CreateFile (dev, GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL)) == INVALID_HANDLE_VALUE) 
    {
        #ifdef DEBUG_MODE
		fprintf(stderr,"Cannot open drive: %d\n", GetLastError ());
        #endif
		dvd = NULL;
	}
    else
    {
        #ifdef DEBUG_MODE
		fprintf(stderr,"Opened successfully\n");
        #endif
		dvd = (DVD_DRIVE_TYPE *) malloc (sizeof (DVD_DRIVE_TYPE));
		dvd->device = strdup(device);
		dvd->fd = fd;
		dvd_get_drive_info (dvd);
		dvd_assign_functions (dvd);
	}

	return (dvd);
}

/**
 * Assigns the proper memory dump functions to a dvd_drive object, according to vendor, model and other parameters. Actually this scheme probably needs to
 * to be improved, but it is enough for the moment.
 * @param dvd The DVD drive the command should be exectued on.
 */
static void dvd_assign_functions (DVD_DRIVE_TYPE *dvd)
{
	if (strcmp (dvd -> vendor, "HL-DT-ST") == 0 && (strcmp (dvd -> prod_id, "DVD-ROM GDR8162B") == 0 ||
	    strcmp (dvd -> prod_id, "DVD-ROM GDR8163B") == 0 || strcmp (dvd -> prod_id, "DVD-ROM GDR8164B") == 0
        || strcmp (dvd -> prod_id, "DVD-ROM GDR3120L") == 0))
    {
        #ifdef DEBUG_MODE
		fprintf(stderr,"Hitachi MN103-based DVD drive detected, using Hitachi memory dump functions\n");
        #endif
		dvd -> memdump_fun_ptr = &hitachi_dvd_dump_mem;
		dvd -> supported = TRUE;
	} else {
		/* This is an unsupported drive (yet). Anyway fall back to Hitachi functions, as we only have those, for the moment, warning the user. */
		#ifdef DEBUG_MODE 
        fprintf(stderr,
			"Unsupported DVD drive model detected, falling back to Hitachi memory\n"
			"dump functions. If the dumping process succeeds, please report the\n"
			"drive identification strings so that it can be added to the supported\n"
			"drives list.\n"
		);
        #endif
        
		dvd -> memdump_fun_ptr = &hitachi_dvd_dump_mem;
		dvd -> supported = FALSE;
	}

	return;
}

/**
 * Sends an INQUIRY command to the drive to retrieve drive identification strings.
 * @param dvd The DVD drive the command should be exectued on.
 * @return 0 if the command was executed successfully, < 0 otherwise.
 */
static S32 dvd_get_drive_info (DVD_DRIVE_TYPE *dvd)
{
	MMC_COMMAND_TYPE mmc;
	S32 out;
	U8 buf[36];
	char tmp[36 * 4];
	
    /* clear mmc struct and data buffer ready for dvd command */
	dvd_init_command (&mmc, buf, sizeof (buf), NULL);

    /* set mmc command type */
	mmc.cmd[0] = SPC_INQUIRY;
	mmc.cmd[4] = sizeof (buf);

	if ((out = dvd_execute_cmd (dvd, &mmc, FALSE)) >= 0)
    {
        dvd->vendor = (char*) calloc(8 + 1,sizeof(char));
        strncpy(dvd->vendor, buf + 8, 8 );
		strtrimr(dvd->vendor);
        dvd->prod_id = (char*) calloc(16 + 1,sizeof(char));
        strncpy(dvd->prod_id, buf + 16, 16);
		strtrimr (dvd -> prod_id);
        dvd->prod_rev = (char*) calloc(4 + 1,sizeof(char));
        strncpy(dvd->prod_rev, buf + 32, 4);
		strtrimr(dvd -> prod_rev);
		snprintf (tmp, sizeof (tmp), "%s/%s/%s", dvd -> vendor, dvd -> prod_id, dvd -> prod_rev);
		dvd->model_string = strdup(tmp);
		
        #ifdef DEBUG_MODE
		fprintf(stderr,"DVD drive is \"%s\"\n", dvd->model_string);
        #endif
	}
    else
    {
        #ifdef DEBUG_MODE
		fprintf(stderr,"Cannot identify DVD drive\n");
        #endif
	}

	return (out);
}

/**
 * Initializes a structure representing an MMC command.
 * @param mmc A pointer to the MMC command structure.
 * @param buf The buffer where results of the MMC command execution provided by the drive should be stored, or NULL if no buffer will be provided.
 * @param len The length of the buffer (ignored in case buf is NULL).
 * @param sense A pointer to a structure which will hold the SENSE DATA got from the drive after the command has been executed, or NULL.
 */
void dvd_init_command (MMC_COMMAND_TYPE *mmc, U8 *buf, S32 len, REQ_SENSE_TYPE *sense)
{
	memset (mmc, 0, sizeof (MMC_COMMAND_TYPE));

	if(buf != NULL)
    {
		memset (buf, 0, len);
    }

	mmc -> buffer = buf;
	mmc -> buflen = buf ? len : 0;
	mmc -> sense = sense;
	
	return;
}

/**
 * Executes an MMC command.
 * @param dvd The DVD drive the command should be exectued on.
 * @param mmc The command to be executed.
 * @param ignore_errors If set to true, no error will be printed if the command fails.
 * @return 0 if the command was executed successfully, < 0 otherwise.
 */
S32 dvd_execute_cmd (DVD_DRIVE_TYPE *dvd, MMC_COMMAND_TYPE *mmc, U8 ignore_errors)
{
	SCSI_PASS_THROUGH_DIRECT *sptd;
	U8 sptd_sense[sizeof (*sptd) + 18];
    U8 *sense;
	DWORD bytes;
	S32 out;

	sptd = (SCSI_PASS_THROUGH_DIRECT *) sptd_sense;
	sense = &sptd_sense[sizeof (*sptd)];
	
	memset (sptd, 0, sizeof (sptd_sense));
	memcpy (sptd -> Cdb, mmc -> cmd, sizeof (mmc -> cmd));
	sptd -> Length = sizeof (SCSI_PASS_THROUGH);
	sptd -> CdbLength = 12;
	sptd -> SenseInfoLength = 18;
	sptd -> DataIn = SCSI_IOCTL_DATA_IN;
	sptd -> DataBuffer = mmc -> buffer;

	/* Quick hack: Windows hates sptd->DataTransferLength = 1, so we set it to 2 and ignore the second byte. */
	if (mmc -> buflen == 1)
    {
		sptd -> DataTransferLength = 2;
    }
	else
    {
		sptd -> DataTransferLength = mmc -> buflen;
    }

	sptd -> TimeOutValue = MMC_CMD_TIMEOUT;
	sptd -> SenseInfoOffset = sizeof (*sptd);

	if (!DeviceIoControl (dvd -> fd, IOCTL_SCSI_PASS_THROUGH_DIRECT, sptd, sizeof (*sptd) + 18, sptd, sizeof (*sptd) + 18, &bytes, NULL) && !ignore_errors)
    {
		out = -1;	/* Failure */
		fprintf(stderr,"Execution of MMC command failed: %s", strerror (errno));
		fprintf(stderr,"Command was: %x length: %d", mmc -> cmd, sizeof (mmc -> cmd));
		fprintf(stderr,"Sense data: %02X/%02X/%02X\n", sense[2] & 0x0F, sense[12], sense[13]);
	}
    else
    {
		out = 0;
	}
	
	if (mmc -> sense)
    {
		mmc -> sense -> sense_key = sense[2];
		mmc -> sense -> asc = sense[12];
		mmc -> sense -> ascq = sense[13];
	}
	
	return (out);
}

/**
 * Executes the drive-dependent function to dump the drive sector cache, and returns the dumped data.
 * @param dvd The DVD drive the command should be exectued on.
 * @param block_off The offset to start dumping, WRT the beginning of the sector cache.
 * @param block_len The number of blocks to dump.
 * @param block_size The block size to be used for dumping.
 * @param buf A buffer where to store the dumped data. Note that this must be able to hold at least block_len * block_size bytes.
 * @return 0 if the command was executed successfully, < 0 otherwise.
 */
S32 dvd_memdump (DVD_DRIVE_TYPE *dvd, U32 block_off, U32 block_len, U32 block_size, U8 *buf)
{
	S32 out;

	out = dvd->memdump_fun_ptr(dvd, block_off, block_len, block_size, buf);
	
    return(out);
}

/**
 * Issues a READ(12) command without bothering to return the results. Uses the FUA (Force Unit Access bit) so that the requested sectors are actually read
 * at the beginning of the cache and can be dumped later.
 * @param dvd The DVD drive the command should be exectued on.
 * @param sector The sector to be read. What will be cached is the 16-sectors block to which the sector belongs.
 * @param sense A pointer to a structure which will hold the SENSE DATA got from the drive after the command has been executed.
 * @return 0 if the command was executed successfully, < 0 otherwise.
 */
S32 dvd_read_sector_dummy (DVD_DRIVE_TYPE *dvd, U32 sector, REQ_SENSE_TYPE *sense)
{
	MMC_COMMAND_TYPE mmc;
	S32 out;
	U8 buf[2048 * 16];
	
	dvd_init_command (&mmc, buf, sizeof (buf), sense);

	mmc.cmd[0] = MMC_READ_12;
	mmc.cmd[1] = 0x08;	/* FUA bit set */
	mmc.cmd[2] = (U8) ((sector & 0xFF000000) >> 24);	/* LBA from MSB to LSB */
	mmc.cmd[3] = (U8) ((sector & 0x00FF0000) >> 16);
	mmc.cmd[4] = (U8) ((sector & 0x0000FF00) >> 8);
	mmc.cmd[5] = (U8) (sector & 0x000000FF);
	mmc.cmd[6] = (((U32) sizeof (buf)) & 0xFF000000) >> 24;
	mmc.cmd[7] = (((U32) sizeof (buf)) & 0x00FF0000) >> 16;
	mmc.cmd[8] = (((U32) sizeof (buf)) & 0x0000FF00) >> 8;
	mmc.cmd[9] = ((U32) sizeof (buf)) & 0x0000000FF;
	out = dvd_execute_cmd (dvd, &mmc, TRUE);		/* Ignore errors! */
	
	return (out);
}

/**
 * Issues a READ(12) command using the STREAMING bit, which causes the requested 16-sector block to be read into memory,
 * together with the following four. This way we will be able to dump 5 sector with a single READ request.
 *
 * Note the strange need for a big buffer even though we must only pass 0x10 as the transfer length, otherwise the drive will hang (!?).
 * @param dvd The DVD drive the command should be exectued on.
 * @param sector The sector to be read. What will be cached is the 16-sectors block to which the sector belongs, and the following 4 blocks.
 * @param sense A pointer to a structure which will hold the SENSE DATA got from the drive after the command has been executed.
 * @param extbuf A buffer where to store the read data, or NULL.
 * @param extbufsize The size of the buffer.
 * @return 
 */
S32 dvd_read_sector_streaming (DVD_DRIVE_TYPE *dvd, U32 sector, REQ_SENSE_TYPE *sense, U8 *extbuf, size_t extbufsize)
{
	MMC_COMMAND_TYPE mmc;
	S32 out;
	U8 intbuf[2048 * 16], *buf;
	size_t bufsize;

	/* We need some buffer, be it provided externally or not */
	if (extbuf)
    {
		buf = extbuf;
		bufsize = extbufsize;
	} 
    else 
    {
		buf = intbuf;
		bufsize = sizeof (intbuf);
	}
	
	dvd_init_command (&mmc, buf, bufsize, sense);
	mmc.cmd[0] = MMC_READ_12;
	mmc.cmd[2] = (U8) ((sector & 0xFF000000) >> 24);	/* LBA from MSB to LSB */
	mmc.cmd[3] = (U8) ((sector & 0x00FF0000) >> 16);
	mmc.cmd[4] = (U8) ((sector & 0x0000FF00) >> 8);
	mmc.cmd[5] = (U8) (sector & 0x000000FF);
	mmc.cmd[6] = 0;
	mmc.cmd[7] = 0;
	mmc.cmd[8] = 0;
	mmc.cmd[9] = 0x10;
	mmc.cmd[10] = 0x80;	/* STREAMING bit set */
	out = dvd_execute_cmd (dvd, &mmc, TRUE);		/* Ignore errors! */
	
	return (out);
}

/**
 * Frees resources used by a DVD drive structure and destroys it.
 * @param dvd The DVD drive structure to be destroyed.
 * @return NULL.
 */
void *dvd_drive_destroy (DVD_DRIVE_TYPE *dvd)
{
	if (dvd)
    {

		CloseHandle (dvd -> fd);

		free(dvd -> device);
		free(dvd -> vendor);
		free(dvd -> prod_id);
		free(dvd -> prod_rev);
		free(dvd);
	}

	return (NULL);
}

char *dvd_get_vendor (DVD_DRIVE_TYPE *dvd)
{
	return (dvd -> vendor);
}


char *dvd_get_product_id (DVD_DRIVE_TYPE *dvd)
{
	return (dvd -> prod_id);
}


char *dvd_get_product_revision (DVD_DRIVE_TYPE *dvd)
{
	return (dvd -> prod_rev);
}


char *dvd_get_model_string (DVD_DRIVE_TYPE *dvd)
{
	return (dvd -> model_string);
}


U8 dvd_get_support_status (DVD_DRIVE_TYPE *dvd) 
{
	return (dvd -> supported);
}
