#include <stdio.h>
#include <windows.h>
#include "types.h"
#include "dvd_drive.h"

/*! \brief Memory offset at which the drive cache memory is mapped.
 *
 * During READ, sector data is cached beginning at this address.
 */

#define HITACHI_MEM_BASE 0x80000000

static S32 hitachi_dvd_dump_memblock (DVD_DRIVE_TYPE *dvd, U32 offset, U32 block_size, U8 *buf);

/**
 * Dumps a given portion of the address space of the MN103 microcontroller within the Hitachi-LG Xbox 360 DVD drive, starting at the offset at which
 * sector data are cached.
 * WARNING: it can take a while to dump a lot of data.
 * @param dvd The DVD drive the command should be exectued on.
 * @param offset The memory offset to start dumping, relative to the cache start offset.
 * @param block_len The number of blocks to dump.
 * @param block_size The block size for the dump.
 * @param buf Where to place the dumped data. This must have been setup by the caller to store up to block_size bytes.
 * @return < 0 if an error occurred, 0 otherwise.
 */
S32 hitachi_dvd_dump_mem (DVD_DRIVE_TYPE *dvd, U32 offset, U32 block_len, U32 block_size, U8 *buf)
{
	U32 i;
	S32 result;
    S32 out;

	if (!buf)
    {
		printf("NULL buffer\n");
		out = -1;
	} 
    else
    {
		result = -10;

		for (i = 0; i < block_len; i++)
        {
			if ((result = hitachi_dvd_dump_memblock (dvd, HITACHI_MEM_BASE + offset + i * block_size, block_size, buf + i * block_size)) < 0)
            {
				printf("hitachi_dvd_read_block() failed with %d", result);
				break;
			}
		}
		out = result;
	}

	return (out);
}

/**
 * Dumps a single block (with arbitrary size) of the address space of the MN103 microcontroller within the Hitachi-LG Xbox 360 DVD drive and similar drives.
 * This function is derived from the work of Kevin East (SeventhSon), kev@kev.nu, http://www.kev.nu/360/.
 * @param dvd The DVD drive the command should be exectued on.
 * @param offset The absolute memory offset to start dumping.
 * @param block_size The block size for the dump.
 * @param buf Where to place the dumped data. This must have been setup by the caller to store up to block_size bytes.
 * @return < 0 if an error occurred, 0 otherwise.
 */
static S32 hitachi_dvd_dump_memblock (DVD_DRIVE_TYPE *dvd, U32 offset, U32 block_size, U8 *buf)
{
	MMC_COMMAND_TYPE mmc;
	S32 out;

	if (!buf)
    {
		printf ("NULL buffer\n");
		out = -1;
	}
    else if (!block_size || block_size > 65535)
    {
		printf ("invalid block_size (valid: 1 - 65535)\n");
		out = -2;
	}
    else
    {
		dvd_init_command (&mmc, buf, block_size, NULL);
		mmc.cmd[0] = 0xE7; // vendor specific command (discovered by DaveX)
		mmc.cmd[1] = 0x48; // H
		mmc.cmd[2] = 0x49; // I
		mmc.cmd[3] = 0x54; // T
		mmc.cmd[4] = 0x01; // read MCU memory sub-command
		mmc.cmd[6] = (unsigned char) ((offset & 0xFF000000) >> 24);	// address MSB
		mmc.cmd[7] = (unsigned char) ((offset & 0x00FF0000) >> 16);	// address
		mmc.cmd[8] = (unsigned char) ((offset & 0x0000FF00) >> 8);	// address
		mmc.cmd[9] = (unsigned char) (offset & 0x000000FF);		// address LSB
		mmc.cmd[10] = (unsigned char) ((block_size & 0xFF00) >> 8);	// length MSB
		mmc.cmd[11] = (unsigned char) (block_size & 0x00FF);		// length LSB

		out = dvd_execute_cmd (dvd, &mmc, FALSE);
	}

	return (out);
}
