#include "stdio.h"
#include "types.h"
#include "ecma-267.h"
#include "constants.h"
#include "byteorder.h"

#define MAX_STREAM_LEN        32767

#define DVD_PRE_SET_OFFSET    0x800
#define GAMECUBE_OFFSET       0x3C00

#define EDC_LENGTH            (RAW_SECTOR_SIZE - 4) 

static U8 unscramble_stream[MAX_STREAM_LEN];
static U8 GameIDChecksum = 0;

static U32 calculate_seed(U32 sector_no, U8 *inbuf);
static U8 unscramble_block (U32 seed, U8 *inbuf, U8 *outbuf);
static U8 unscramble_frame (U32 seed, U8 *inbuf, U8 *outbuf);

/**
 * Create dvd unscrambling stream
 * @return NULL.
 */
void unscramble_init(void)
{
    S32 index;

    /* give the LFSR a seed of one */
    LFSR_init(1);
    unscramble_stream[0] = 1;

    /* Create dvd unscrambling stream */
    for(index = 1 ; index < MAX_STREAM_LEN ;index++)
    {
        unscramble_stream[index] = LFSR_byte();
    }

	return;
}

/**
 * Unscrambles a 16-sector block.
 * @param u The unscrambler structure.
 * @param sector_no The number of the first sector in the block.
 * @param inbuf The 16-sector block to unscramble. Each block must be RAW_SECTOR_SIZE bytes long, so that the total size is RAW_BLOCK_SIZE.
 * @param outbuf The unscrambled 16-sector block. Each block will be SECTOR_SIZE bytes long, so that the total size is BLOCK_SIZE.
 * @return True if the unscrambling was successful, false otherwise.
 */
U8 unscramble_16sectors (U32 sector_no, U8 *inbuf, U8 *outbuf)
{
    U32 seed_index;
	U8 out = FALSE;

	/* get seed index value based on sector number */
    seed_index = calculate_seed(sector_no, inbuf);
	
	/* try and unscramble block with computed seed */
	if(0 == unscramble_block (seed_index, inbuf, outbuf))
	{
        #ifdef DEBUG_MODE
	    fprintf (stderr,"Error unscrambling block %u seed %04x \n", sector_no,seed_index);
        #endif
	    out = FALSE;
	}
	else
	{
	    out = TRUE;
	}
	
	return (out);
}

void rescramble_16sectors(U32 sector_no, U8 *inbuf)
{
    U32 i;
    U32 j;
    U32 k;
    U8 *bin;
    U32 seed_index; 

    seed_index = ((sector_no & 0x000000F0) >> 4) * DVD_PRE_SET_OFFSET;

    for(j = 0; j < 16; j++)
    {
        bin = &inbuf[RAW_SECTOR_SIZE * j];

        for(i = 12, k = seed_index; i < EDC_LENGTH; i++, k++)
        {
            if(k == MAX_STREAM_LEN)
            {
                k = 0;
            }
           
            bin[i] ^=  unscramble_stream[k];
        }
    }
}

/* directly set the gameid checksum */
void unscramble_set_gameid(U8 id)
{
    GameIDChecksum = id;
}

static U32 calculate_seed(U32 sector_no, U8 *inbuf) 
{
    U32 seed_index = 0; 

    /* check for first sector block on disk */
    if(sector_no == 0)
    {
        /* for sector 0 just use the gamecube seed offset only */
        seed_index = GAMECUBE_OFFSET;  
    }
    else
    {
        /* use gamecube seed index calculation */
        seed_index = ((((sector_no & 0x000000F0) >> 4) ^ GameIDChecksum ) * DVD_PRE_SET_OFFSET ) + GAMECUBE_OFFSET;
    }

    /* make sure seed_index stays within array bounds */
    if(seed_index > MAX_STREAM_LEN)
    {
        seed_index =  seed_index % MAX_STREAM_LEN;
    }

    return (seed_index);
}

static U8 unscramble_block (U32 seed, U8 *inbuf, U8 *outbuf)
{
    U8 *bin;
    U8 *bout;
    U8 out;
    U32 j;

    out = TRUE;

    for(j = 0; j < 16; j++)
    {
        bin = &inbuf[RAW_SECTOR_SIZE * j];
        bout = &outbuf[SECTOR_SIZE * j];
        if(0 == unscramble_frame (seed, bin, bout))
        {
            out = FALSE;
            break;
        }
    }
}

static U8 unscramble_frame (U32 seed, U8 *inbuf, U8 *outbuf)
{
    U32 edc_calculated;
    U32 edc_correct;
    U8 out;
    U32 i;
    U32 j;
    U8 tmp[RAW_SECTOR_SIZE];

    memcpy (tmp, inbuf, RAW_SECTOR_SIZE);

    /* unscramble frame */
    for (i = 12, j = seed ; i < EDC_LENGTH; i++, j++)
    {
        /* wrap stream around if required */
        if(j == MAX_STREAM_LEN)
        {
            j = 0;
        }        
        
		tmp[i] ^= unscramble_stream[j];
    }

	edc_calculated = edc_calc (0x00000000, tmp, EDC_LENGTH);
	edc_correct = my_ntohl (*((U32 *) (&tmp[EDC_LENGTH])));
	
    if (edc_calculated == edc_correct) 
    {
        /* copy out unscrambled data and return */
        memcpy (outbuf, &tmp[6], SECTOR_SIZE);
        out = TRUE;
    }
	else
	{
        #ifdef DEBUG_MODE
        fprintf (stderr,"\nEDC check failed cal'ed value was %8x real value %8x\n",edc_calculated,edc_correct);
        #endif
	    out = FALSE;
	}
	
	return out;
}
